# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import copy
import re
from contextlib import ExitStack
from multiprocessing import Pool
from multiprocessing import cpu_count

import graphviz as gv
from IPython.display import SVG
from IPython.display import display
from graphviz import Source
from tqdm import tqdm

from . import config


def _cluster_run(data_fetcher, stage_containers, run_id=-1, verbose = False):
    ''' 
    Run the pipeline

    @param data_fetcher: Data fetcher to use as a data source (from skdaccess)
    @param stage_containers: List of stage containers
    @param run_id: Run_id to use, for offloading
    @param verbose: Display the tqdm progress bar for this run

    @return results from pipeline run
    '''
    data_container = data_fetcher.output()
    data_container.run_id = run_id

    if verbose == False:
        iterator = stage_containers
    else:
        iterator = tqdm(stage_containers)
    
    for s in iterator:
        s.run(data_container)

    return data_container.getResults()

def _wrap_cluster(args):
    ''' Wrap cluster run 

    @param args: Arguments to pass to _cluster_run

    @return results from _cluster_run
    '''
    return _cluster_run(*args)


class DiscoveryPipeline:
    ''' Pipeline for running the analysis '''
    
    def __init__(self, data_fetcher, list_StageContainers):
        '''
        Initialize a new pipeline.
        
        @param data_fetcher: Data fetcher to use as a data source (from skdaccess)
        @param list_StageContainers: List of stage containers
        '''
        
        self.stage_containers = list_StageContainers
        self.data_fetcher   = data_fetcher
        self.stageConfigurationHistory = []
        self.RA_results = []
        self.__cluster = None
        self._dispy_http = None        

    def run(self, num_runs=1, perturb_data = False, num_cores = 1, amazon=False, verbose=False):
        '''
        Run the pipeline

        @param num_runs: Number of times to run the pipeline
        @param perturb_data: Boolean flag. If running the pipeline multiple times then
                            perturb the data instead of the pipeline
        @param num_cores: Number of cores on the local machine to use. Defaults
                          to 1 core. Use 0 to select the minimum between the
                          number of runs and cpu cores.
        @param amazon: Offload the pipeline on amazon
        @param verbose: Display the pipeline for each run
        '''


        # Check to make sure data fetcher supports amazon or multicore if requested
        if (amazon == True or num_cores != 1) and self.data_fetcher.multirun_enabled() == False:
            raise RuntimeError('Data Fetcher type does not support amazon or using multiple cores')

        # Run the job on Amazon
        if amazon == True:

            if self.__cluster == None:
                self._startCluster(config.getDispyPassword())

            self._createDispyLink()

            # Run Jobs on Amazon
            jobs = []

            for i in range(0,num_runs):
                if num_runs > 1:
                    run_index = i
                else:
                    run_index = -1

                if verbose:
                    self.plotPipelineInstance()
                job = self.__cluster.submit(copy.deepcopy(self.data_fetcher), copy.deepcopy(self.stage_containers), run_id = run_index)
                job.id = i
                jobs.append(job)
                # save metadata configuration for history
                self.stageConfigurationHistory.append(self.getMetadata())

                if num_runs > 1:
                    if perturb_data == False:
                        self.perturb()
                    else:
                        self.perturbData()


            for job in jobs:
                results = job()
                # save results in the result accumulator
                if(job.exception is not None):
                    print(job.exception)
                self.RA_results.append(results)

            self.__cluster.stats()
            self.__cluster.close()
            self.__cluster = None

        # Run the job on the local machine
        else:

            # Function to generate inputs for running
            def generatePipelineInputs():
                self.stageConfigurationHistory.append(self.getMetadata())
                if verbose:
                    self.plotPipelineInstance()
                yield copy.deepcopy(self.data_fetcher), copy.deepcopy(self.stage_containers), 0

                for i in range(1, num_runs):
                    if perturb_data == False:
                        self.perturb()
                    else:
                        self.perturbData()

                    self.stageConfigurationHistory.append(self.getMetadata()) 
                    if verbose:
                        self.plotPipelineInstance()
                    yield copy.deepcopy(self.data_fetcher), copy.deepcopy(self.stage_containers), i

                # Perturb data as many times as runing more than once
                if num_runs > 1:
                    if perturb_data == False:
                        self.perturb()
                    else:
                        self.perturbData()
                    
            # Run the jobs
            with ExitStack() as stack:

                # Create workers for processing multiple pipelines locally
                if num_runs != 1:
                    if num_cores == 0:
                        # Create a pool with either the number of cores on
                        # the machine, or the number of runs, whichever is
                        # less
                        pool = Pool(min(cpu_count(), num_runs))
                        stack.callback(pool.close)
                    elif num_cores > 1:
                        # If number of cores is specified generate that many
                        # workers
                        pool = Pool(num_cores)
                        stack.callback(pool.close)

                if (num_cores == 0 or num_cores > 1) and num_runs != 1:
                    results = pool.map(_wrap_cluster, generatePipelineInputs())
                else:
                    results = list(map(_wrap_cluster, generatePipelineInputs()))

                self.RA_results += results                
        
    def perturb(self):
        ''' Perturb the paramters in the stage containers '''
        for s in self.stage_containers:
            s.perturb()

    def reset(self):
        ''' 
        Reset the stage containers to their default values and clear previous runs.
        '''
        for s in self.stage_containers:
            s.reset()

        self.stageConfigurationHistory = []
        self.RA_results = []

    def getMetadata(self):
        '''
        Retrieve Metadata from stage containers

        @return list of metadata for the current run
        '''

        data_metadata = self.data_fetcher.getMetadata()
        
        metadataList = [s.getMetadata() for s in self.stage_containers]
        returnList = [data_metadata]
        for info  in metadataList:
            if isinstance(info, list):
                returnList += info
            else:
                returnList.append(info)
                
        return returnList

    def getMetadataHistory(self):
        ''' 
        Get the metadata for each run in the pipeline 

        @return list of metadata configurations for all runs
        '''
        return self.stageConfigurationHistory

    def perturbData(self):
        ''' Perturb the input data '''
        self.data_fetcher.perturb()

    def getResults(self,index=None):
        '''
        Return results from previous runs.

        @param index: Index of run. If None, return all previous results

        @return results from a run at index. If index=None, returns list of all results
        '''
        if index==None:
            return self.RA_results
        else:
            return self.RA_results[index]

    def resultIter(self):
        ''' 
        Retrieves and iterator to the results and history of the pipeline.

        @return A 2 component iterator to the results and history of previous runs
        '''
        return zip(self.getResults(), self.getMetadataHistory())

    def plotPipelineInstance(self):
        '''
        Plot current instance of pipeline stages with metadata

        @return iPython display object
        '''
            
        #Graph setup
        g1 = gv.Digraph(format='svg')
        g1.graph_attr['rankdir'] = 'LR'
        g1.node_attr['shape'] = 'rounded'
        g1.node_attr['fontname'] = 'Arial'
        g1.node_attr['fontsize'] = '9'
        g1.node_attr['style'] = 'filled'
        g1.node_attr['margin'] = '0.1'
        g1.node_attr['height'] = '0.1'
        g1.node_attr['fillcolor'] = '#d8e9fd'
        #g1.node_attr['shape'] = 'plaintext'  #use this to remove boxes around nodes

        # Some stagecontainers directly return the metadata as string,
        # others return a list of strings for each item in the stage
        # container.
        metadata_list = []
        for s in self.stage_containers:
            first_metadata = s.getMetadata()
            if isinstance(first_metadata, str):
                metadata_list.append(first_metadata)

            else:
                for second_metadata in first_metadata:
                    metadata_list.append(second_metadata)
        
        nodelist = [re.sub('\:',';', metadata) for metadata in metadata_list]

        for s in nodelist:
            g1.node(s)
        
        g1.edges(zip(nodelist, nodelist[1:]))
        
        g1.render('img/plotPipelineInstance')
        
        #print(g1.source)
        #return display(Image('img/pipelinePlot.svg'))
        return display(SVG('img/plotPipelineInstance.svg'))

    def plotPipelineStructure(self):
        '''
        Plot pipeline structure

        @return iPython display object
        '''
            
        #Graph setup
        g1 = gv.Digraph(format='svg')
        g1.graph_attr['rankdir'] = 'LR'
        g1.node_attr['shape'] = 'rounded'
        g1.node_attr['fontname'] = 'Arial'
        g1.node_attr['fontsize'] = '9'
        g1.node_attr['style'] = 'filled'
        g1.node_attr['margin'] = '0.1'
        g1.node_attr['height'] = '0.1'
        g1.node_attr['fillcolor'] = '#fff7da'
        #g1.node_attr['shape'] = 'plaintext'  #use this to remove boxes around nodes
        
        nodelist= self.getMetadataNestedGraph()
        
        print (nodelist)
        src = Source(nodelist)
        print (dir(src))
        src.format='svg'
        src.render('img/plotPipelineStructure')
        
#        for s in nodelist:
#            g1.node(s)
#        g1.edges(zip(nodelist, nodelist[1:]))
#        g1.render('img/plotPipelineStructure')       
#        print(nodelist)
#        print(g1.source)
#        return display(SVG('img/plotPipelineStructure.svg'))

        return display(SVG('img/plotPipelineStructure.svg'))

    #++++++++++ NEW Victor +++++++++++
    def getMetadataNestedTypes(self):
        '''
        Get the Metadata Nested Types

        @return String: Metadata Nested types
        '''
        returnString=""

        for e in self.stage_containers:
            returnString = returnString + e.getMetadataNestedTypes()
            
        return returnString
    
    #++++++++++ NEW Victor +++++++++++
    def getMetadataNestedGraph(self):
        '''
        Retrieve the metadata nested graph

        @return String: Metadata nested graph
        '''
        returnString="digraph { \n graph [rankdir=LR] \n node [fillcolor=\"#d8e9fd\" fontname=Arial fontsize=9 height=0.1 margin=0.1 shape=rounded style=filled]\n"

        for e in self.stage_containers:
            returnString = returnString + e.getMetadataNestedGraph()
        
        returnString = returnString + "}"
        
        return returnString


    def _startCluster(self, secret):
        ''' 
        Start dispy cluster

        @param secret: Password for dispy nodes
        '''

        import dispy
        
        # Amazon run function
        def amazon_run(data_fetcher, stage_containers, run_id=-1):
            data_container = data_fetcher.output()
            data_container.run_id = run_id
            for s in stage_containers:
                s.run(data_container)

            return data_container.getResults()

        self.__cluster = dispy.SharedJobCluster(amazon_run, secret=secret, port=0,
                                                scheduler_node='127.0.0.1',
                                                ip_addr='127.0.0.1')

        # for i in range(1000):
        #     try:
        #         self._dispy_http = DispyHTTPServer(self.__cluster,port = 8181 + i)
        #         self._dispy_http.port = 8181 + i
        #         break
        #     except OSError:
        #         pass

            
    def _createDispyLink(self):
        ''' Create a link to the Dispy Cluster Information website '''
        
        print("Access cluster monitor at http://hostname:8181 where hostname is the address of the primary host")
        
    def __str__(self):
        '''
        String representation of the pipeline.

        @return String of current metadata of pipeline containers.
        '''
        return str(self.getMetadata())



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import numpy as np
from sklearn.decomposition import PCA
from sklearn.decomposition import FastICA



class General_Component_Analysis(PipelineItem):
    ''' 
    Performs either ICA or PCA analysis on series data
    '''

    def __init__(self, str_description, ap_paramList):
        '''
        Initialize Analysis object.

        @param str_description: String description of analysis
        @param ap_paramList[num_components]: Number of components
        @param ap_paramList[component_type]: Type of component analysis (CA); either PCA or ICA
        @param ap_paramList[start_time]: Starting time for CA
        @param ap_paramList[end_time]: ending time for CA
        @param ap_paramList[labels]: Optional list of label names
        '''
        
        self.str_description = str_description
        self.ap_paramList = ap_paramList
        self.ap_paramNames = ['n_components','component_type','start_time','end_time','label_names']
        self.results = dict()

    
    def process(self, obj_data):
        ''' 
        Perform component analysis on data:

        Results are added to the data wrapper as a dictionary with
        results['CA'] = Eigenvenctors
        results['Projection'] = Projection on to the eigenvectors
        
        @param obj_data: Data wrapper containing the data
        '''

        num_components = self.ap_paramList[0]()
        component_type = self.ap_paramList[1]()
        start_time = self.ap_paramList[2]()
        end_time = self.ap_paramList[3]()

        results = dict()
        results['start_date'] = start_time
        results['end_date'] = end_time

        if len(self.ap_paramList) >= 5:
            label_names = self.ap_paramList[4]()
        else:
            label_names = None
            
        cut_data = []
        for label, data, err in obj_data.getIterator():
            if label_names == None or label in label_names:
                cut_data.append(data[start_time:end_time])

        cut_data = np.array(cut_data)

        if len(cut_data) > 0:
            if component_type == 'ICA' :
                ca = FastICA(n_components = num_components)
            else:
                ca = PCA(n_components = num_components)
            
            time_projection = ca.fit_transform(cut_data.T)
            results['CA'] = ca
            results['Projection'] = time_projection

        else:
            results['CA'] = None
            results['Projection'] = None

        obj_data.addResult(self.str_description, results)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem

import numpy as np
import pandas as pd
from sklearn.tree import DecisionTreeRegressor

class OffsetDetrend(PipelineItem):
    '''
    Trend filter that fits a stepwise function to linearly detrended table data

    On detrended data this filter fits a stepwise function (number of
    steps provided by the user) to correct the linear fit by
    accounting for discontinuous offsets, such as due to a change in
    the antenna or from an earthquake. The final linear fit handles
    each portion of the offset independently. If the number of
    discontinuities is not provided as an autoparam, the filter
    assumes a single discontinuity.
    '''

    def __init__(self, str_description, column_names, ap_paramList = [], labels=None, time_point=None, time_interval=None):
        ''' 
        Initialize OffsetDetrend filter for use on table data
        
        @param str_description: String describing filter
        @param column_names: List of column names to select data to be removed (using None will apply to all columns)
        @param ap_paramList[step_count]: Number of steps to remove from data (Default: 1)
        @param labels: List of labels used to select data to be removed (using None will apply to all labels)
        @param time_point: Time of offset
        @param time_interval: Interval within which the offset occurs
        '''
        
        self.labels = labels
        self.column_names = column_names
        self.time_point = time_point
        if time_interval == None:
            self.time_interval = [-500,500]
        else:
            if type(time_interval) == int:
                self.time_interval = [-time_interval,time_interval]        
            else:
                self.time_interval = time_interval
        self.ap_paramNames = ['step_count']
            
        super(OffsetDetrend, self).__init__(str_description, ap_paramList)

    
    def process(self, obj_data):
        ''' 
        Apply offset estimation and detrending filter to data set.
        
        @param obj_data: Input data. Changes are made in place.
        '''

        labels = self.labels
        column_names = self.column_names

        # user provided number of steps/offsets in the data
        step_count = 1
        if len(self.ap_paramList) != 0:
            step_count = self.ap_paramList[0]()
            
        for label, data in obj_data.getIterator():
            for column in column_names:
                if (labels is None or label in labels):
                    # keep track of the time index and the location of nan's                
                    tindex = data.index
                    reind = np.array(np.isnan(data))
                    # a temporary time index and data array without nan's
                    nts = np.arange(len(data))
                    nts = np.delete(nts,nts[reind])
                    nys = data[reind==False]

                    # Decision Tree Regressor for finding the discontinuities
                    regr_1 = DecisionTreeRegressor(max_depth=step_count)
                    if self.time_point == None:
                        regr_1.fit(nts[:,np.newaxis], nys)
                    else:
                        # make time_point (a string) into an index
                        time_point = np.where(tindex==self.time_point)[0][0]
                        regr_1.fit(nts[(time_point+self.time_interval[0]):(time_point+self.time_interval[1]),np.newaxis],
                                   nys[(time_point+self.time_interval[0]):(time_point+self.time_interval[1])])
                    r1 = regr_1.predict(nts[:,np.newaxis])

                    # offset the discontinuity to be continous and fit a single line
                    # (using median of 5 points on either side of discontinuity)
                    nys[r1==r1[-1]] += np.median(nys[r1==r1[0]][-5:-1]) - np.median(nys[r1==r1[-1]][0:5])
                    z3 = np.polyfit(nts, nys, 1)

                    # make the data into a pd series and correctly index
                    x3 = pd.Series(data=nys-(z3[0]*nts+z3[1]),index=tindex[reind==False])
                    x3 = x3.reindex(tindex)
                    # and then use that to update in place
                    obj_data.updateData(label, x3.index, column, x3)
            



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import random
import re
import socket
import string
import subprocess
import time
from collections import OrderedDict

import boto3
import psutil
from paramiko.ssh_exception import SSHException

from skdiscovery.utilities.cloud.ssh_reverse import ReverseTunnel
from skdiscovery.data_structure.framework import config

aws_access_key = None
aws_secret = None
aws_region = None
aws_security_group = None
aws_key_name = None
pem_file = None
ec2_res = None
ec2_client = None
amazon_list = []
scheduler = None
popen = None


def init(in_aws_access_key, in_aws_secret, in_aws_region, in_aws_security_group, in_aws_key_name, in_pem_file):
    '''
    The underlying functionality for the Amazon GUI, the user should not need to directly interface with this function

    @param in_aws_access_key: AWS access key
    @param in_aws_secret: AWS Secret Access Key
    @param in_aws_region: AWS region (e.g. us-west-2)
    @param in_aws_security_group: Security Group Name
    @param in_aws_key_name: Name of Key Pair
    @param in_pem_file: Filename of ssh key
    '''    

    global aws_access_key
    global aws_secret
    global aws_region
    global aws_security_group
    global aws_key_name
    global pem_file
    global ec2_res
    global ec2_client
    global amazon_list
    
    aws_access_key = in_aws_access_key
    aws_secret = in_aws_secret
    aws_region = in_aws_region
    aws_security_group = in_aws_security_group
    aws_key_name = in_aws_key_name
    pem_file = in_pem_file
    ec2_res =  boto3.resource('ec2', aws_access_key_id=aws_access_key, aws_secret_access_key=aws_secret,
                              region_name=aws_region)

    ec2_client = boto3.client('ec2', aws_access_key_id=aws_access_key, aws_secret_access_key=aws_secret,
                              region_name=aws_region)


    clearAmazonList()

    full_info = ec2_client.describe_instances()

    for reservation in full_info['Reservations']:
        for instance in reservation['Instances']:
            if 'Tags' in instance:
                if re.search('^CAD_Cloud_i-[0-9a-zA-Z]*', instance['Tags'][0]['Value']):
                # if re.search('^scale1.*', instance['Tags'][0]['Value']):
                    instance_obj = ec2_res.Instance(id=instance['InstanceId'])
                    new_instance = generateInfo(instance_obj)
                    if new_instance['state'] in ('running','pending'):
                        amazon_list.append(new_instance)

    if len(amazon_list) > 0:
        resetInstances()
        createTunnels()
        startDispyNode()        
        startDispyScheduler()
        

def closeDispyScheduler():
    ''' Close the Dispy Scheduler '''

    global popen
    if popen != None:
        popen.terminate()
        popen.wait()
        popen=None

    else:
        for proc in psutil.process_iter():
            try:
                cmdline = proc.cmdline()
            except (PermissionError, psutil.AccessDenied):
                continue

            for arg in cmdline:
                if re.search('dispyscheduler.py',arg):
                    proc.send_signal(psutil.signal.SIGTERM)
    
def startDispyScheduler():
    ''' Start the Dispy Scheduler '''

    closeDispyScheduler()

    dispy_pass = config.getDispyPassword()    

    global popen
    
    args = ['dispyscheduler.py', '--node_secret', dispy_pass, '--ip_addr',
            '127.0.0.1', '--cluster_secret', dispy_pass, '--pulse_interval',
            '60','--msg_timeout', '60', '--httpd', '-p', '61234', '--daemon']

    for info in amazon_list:
        args.append('-n')
        args.append(info['ip_address'])

    popen =  subprocess.Popen(args, shell=False)
    

def generateInfo(instance):
    '''
    Read metadata from an Amazon instance

    @return metadata for Amazon instance
    '''
    new_info = OrderedDict()
    new_info['amazon_id'] = instance.id
    new_info['state'] = instance.state['Name']
    new_info['ip_address'] = instance.public_ip_address
    new_info['tunnel'] = None
    new_info['instance'] = instance
    return new_info

def updateStatus():
    ''' Update status information in amazon_list '''

    global amazon_list
    id_list = [instance['amazon_id'] for instance in amazon_list]
    status_info = ec2_client.describe_instance_status(InstanceIds = id_list, IncludeAllInstances=True)
    for status in status_info['InstanceStatuses']:
        for info in amazon_list:
            if info['amazon_id'] == status['InstanceId']:
                info['state'] = status['InstanceState']['Name']
                break

def setNumInstances(new_total_instances, instance_type, image_id):
    ''' 
    Change the number of running instances

    @param new_total_instances: New number of instances
    @param instance_type: Instance type for new instances
    @param image_id: ID of image (ami-xxxxxxxx)
    '''

    global amazon_list
    orig_total_instances = len(amazon_list)
    
    delta_instances = new_total_instances - orig_total_instances

    if delta_instances > 0:

        closeDispyScheduler()
        resetInstances()

        new_instances = ec2_res.create_instances(ImageId=image_id, MinCount=delta_instances, 
                                                      MaxCount=delta_instances, InstanceType=instance_type, 
                                                      SecurityGroups=[aws_security_group], KeyName=aws_key_name)

        for instance in new_instances:
            instance.create_tags(Tags=[{'Key':'Name','Value':'CAD_Cloud_' + instance.id}])
            new_info = generateInfo(instance)
            amazon_list.append(new_info)


    elif delta_instances < 0:

        closeDispyScheduler()
        ec2_client.terminate_instances(InstanceIds=[a['amazon_id'] for a in 
                                                         amazon_list[delta_instances:]])

        amazon_list = amazon_list[:new_total_instances]

        if new_total_instances != 0:
            resetInstances()
        

    if new_total_instances not in (orig_total_instances, 0):
        createTunnels()
        startDispyNode()
        startDispyScheduler()


def updateIPAddress(instance_info):
    '''
    Update ip address of instance info

    @param instance_info: Information about amazon instance
    '''
    try:
        information = ec2_client.describe_instances(InstanceIds=[instance_info['amazon_id']])

        instance_info['ip_address'] = information['Reservations'][0]['Instances'][0]['PublicIpAddress']

    except KeyError:
        pass


def goodConnection(instance, port):
    '''
    Check if an amazon instance has a port open

    @param instance: Amazon instance information
    @param port: Port to check

    @return Boolean indicating if a port is open
    '''
    test_socket = socket.socket()

    success = False

    try:
        if instance['ip_address'] == None:
            updateIPAddress(instance)


        if instance['ip_address'] != None:

            test_socket.connect((instance['ip_address'], port))
            test_socket.shutdown(socket.SHUT_RDWR)
            test_socket.close()

            success = True

    except (socket.error, TypeError):
        pass

    return success


def createTunnels():
    ''' Create reverse ssh tunnels to all instances '''

    while len([i for i in amazon_list if i['tunnel'] == None]) > 0:
        time.sleep(5)
        for instance in amazon_list:
            if instance['tunnel'] == None and instance['state'] == 'running' \
               and goodConnection(instance, 22):

                try:
                    instance['tunnel'] = \
                                ReverseTunnel(instance['ip_address'], 'ubuntu',
                                              pem_file, 61234, '127.0.0.1', 61234, check=30)
                    instance['tunnel'].create_reverse_tunnel()
                except SSHException:
                    print('cannot create tunnel')
                    instance['tunnel'] = 'NA'

        updateStatus()

    time.sleep(10)

def startDispyNode():
    ''' Start dispy on each Amazon instance '''
    
    dispy_pass = config.getDispyPassword()

    if dispy_pass == None:
        dispy_pass = ''.join([random.choice(string.ascii_letters + string.digits)
                              for x in range(20)])
        config.writeConfigValue('Dispy', 'password', dispy_pass)

    command = 'bash -c "export PATH=/home/ubuntu/miniconda3/bin:$PATH; source activate py36; /home/ubuntu/dispy/ssh_dispy.sh'


    for instance in amazon_list:
        stdin, stdout, stderr = instance['tunnel'].ssh.exec_command(command + " '" + dispy_pass + "'\"")

    # Need to give time for dispynode to start
    time.sleep(20)


    index = 0
    while index < len(amazon_list):
        if goodConnection(amazon_list[index], 51348):
            index += 1

        else:
            time.sleep(10)

    time.sleep(10)


def resetInstances():
    '''
    Reboot Amazon instances
    '''
    global amazon_list

    instanceids = [instance['amazon_id'] for instance in amazon_list]

    if len(amazon_list) > 0:
        ec2_client.reboot_instances(InstanceIds = instanceids)

        for instance in amazon_list:
            if instance['tunnel'] != None:
                if instance['tunnel'] != 'NA':
                    instance['tunnel'].event.set()
                    del instance['tunnel']
                    
                instance['tunnel'] = None
    
def reset():
    ''' Close and clear Amazon List '''
    closeDispyScheduler()
    
    clearAmazonList()
        
def close():
    ''' Shutdown all instances, close dispy scheduler and clear Amazon list '''

    setNumInstances(0)        
    closeDispyScheduler()

    clearAmazonList()    


def clearAmazonList():
    ''' Shutdown connection tunnels to Amazon instances and clear amazon list '''
    global amazon_list


    for instance in amazon_list:
        if instance['tunnel'] != None:
            instance['tunnel'].event.set()
            del instance['tunnel']
            instance['tunnel'] = None
    

    amazon_list = []



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from skdiscovery.data_structure.framework.base import PipelineItem
from skdiscovery.data_structure.framework.stagecontainers import *
from skdaccess.framework.param_class import *
from skdaccess.geo.imsdnhs import DataFetcher as SDF
from collections import OrderedDict


class SnowFusion(PipelineItem):
    ''' 
    Adds snow time series data to table based on geographic coordinates 

    Works on table data (original data from http://nsidc.org/data/g02156)
    '''
    def __init__(self, str_description, metadata, column_data_name = 'Snow'):
        '''
        Initialize Snow Fusion item

        @param str_description: String describing item
        @param metadata: Metadata that contains lat,lon coordinates based on data labels
        @param column_data_name: Name of column for Snow data
        '''
        
        super(SnowFusion, self).__init__(str_description, [])
        self.metadata = metadata
        self.column_data_name = column_data_name



    def process(self, obj_data): 
        ''' 
        Adds column for snow (g02156) data

        @param obj_data: Input DataWrapper, will be modified in place
        '''
        coordinate_dict = OrderedDict()
        for label, data in obj_data.getIterator():
            coordinate_dict[label] = (self.metadata[label]['Lat'], self.metadata[label]['Lon'])
            start_date = data.index[0]
            end_date = data.index[-1]
            sdf = SDF(coordinate_dict, start_date, end_date)
            snow = sdf.output().get()[label].loc[:,'Snow']

            obj_data.addColumn(label, self.column_data_name, snow)



from setuptools import setup
from setuptools import find_packages

package_name = 'scikit-discovery'

package_list = find_packages()

with open("README.md", 'r', encoding='utf-8') as rfile:
    readme = rfile.read()

setup(name     = package_name,
      version  = '0.9.14.post1',
      packages = package_list,
      zip_safe = False,

      install_requires = [
          'astropy>=1.1.2',
          'basemap',
          'boto3>=1.4.4',
          'dispy==4.8.1',
          'graphviz >= 0.7.0',
          'ipython >= 6.0.0',
          'ipywidgets >= 6.0.0',
          'matplotlib >= 2',
          'numpy>=1.10',
          'pandas>=0.17',
          'paramiko >= 2.1.0',
          'psutil >= 5.2',
          'pyproj',
          'pyhht',
          'PyWavelets',
          'scikit-dataaccess>=1.9.0',
          'scikit-learn >= 0.17',
          'scipy',
          'seaborn >= 0.8',
          'setuptools',
          'six >= 1.10.0',
          'statsmodels >= 0.8',
          'tqdm',
      ],

      description = 'A package for Computer-Aided Discovery',
      author = 'MITHAGI',
      author_email='skdaccess@mit.edu',
      classifiers=[
          'Topic :: Scientific/Engineering',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
          'Programming Language :: Python :: 3 :: Only'
          ],
      python_requires='>=3.4',
      url = 'https://github.com/MITHaystack/scikit-discovery',
      package_data={'skdiscovery': ['license/LICENSE','license/LGPL_LICENSE','license/MIT_LICENSE',
                                    'docs/skdiscovery_doxygen.pdf','examples/Amazon_GUI.ipynb',
                                    'examples/Amazon_Offload.ipynb']},

      long_description = readme,
      long_description_content_type='text/markdown'
)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem

from skdiscovery.utilities.patterns import trend_tools


class MedianFilter(PipelineItem):
    '''
    A Median filter for series data
    '''

    def __init__(self, str_description, ap_paramList, interpolate=True, subtract = False):
        '''
        Initialize MedianFilter
        
        @param str_description: String describing filter
        @param ap_paramList[ap_window]: median filter window width
        @param interpolate: Flag to interpolate data points before filtering
        @param subtract: Flag to subtract filtered result from original
        '''

        self.interpolate = interpolate
        self.subtract = subtract
        self.ap_paramNames = ['windowSize']
        super(MedianFilter, self).__init__(str_description, ap_paramList)

    
    def process(self, obj_data):
        ''' 
        Apply median filter to data set
        
        @param obj_data: Input DataWrapper. Changes are made in place.
        '''

        ap_window = self.ap_paramList[0]()

        for label, data, err in obj_data.getIterator():

            result = trend_tools.medianFilter(data, ap_window, self.interpolate)
            if self.subtract == True:
                data.update(data - result)
            else:
                data.update(result)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from skdiscovery.data_structure.framework import PipelineItem
from statsmodels.robust import mad


class Outlier(PipelineItem):

    ''' 
    Computes (data / mad(data)) for outlier detection

    Creates a new column for the result
    '''

    def __init__(self, str_description, columns=None, name_prefix = 'MAD_Scale_'):
        '''
        Initalize Outlier Item

        @param str_description: Name of Item
        @param columns: List of of column names
        @param name_prefix: Prefix of newly created column
        '''

        self.columns = columns
        self.name_prefix = name_prefix
        super(Outlier, self).__init__(str_description)

    def process(self, obj_data):
        '''
        Process the data object to add a column with the outlier scores

        @param obj_data: Input table data wrapper
        '''

        if self.columns == None:
                column_names = obj_data.getDefaultColumns()
        else:
                column_names = self.columns

        for label, data in obj_data.getIterator():
            for column in column_names:

                clean_data = data[column].dropna()
                
                mad_scale = clean_data / mad(clean_data)
                obj_data.addColumn(label, self.name_prefix + column, mad_scale)





# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Author: Justin D Li
# This software is part of the NASA AIST Project "Computer-Aided Discovery of Earth Surface Deformation Phenomena" , PI: V. Pankratius
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# '''
# @package trendTools
#   This module is designed to provide a suite of tools for quick analysis of
#   the linear and sinusoidal (annual, semi-annual, seasonal, and monthly)
#   trends of time-series data (formatted using the pandas format).
#'''


import numpy as np
import pandas as pd
from scipy import optimize
from scipy.ndimage import median_filter


def getTrend(xdata):
    '''
    The getTrend function applies the signal.detrend function

    Returns the trend, given a time index input.
    
    @param xdata: 1D time-series data in a pandas series format
    
    @return the detrended data in pandas series format
    @return the linear trend assuming a 1 day per sample time fit
    @return the parameters for the linear trend
    '''
    
    # returns detrended data and the trend
    # checks the time frequency to create a time vector
    if xdata.index.freq != pd.tseries.offsets.Day():
        print('Frequency of Sampling is not Day, please check that resampling occurred!')
    timeDat = np.arange(0,len(xdata.index))
    dataDat = xdata
    # remove time and data with NaN's for now
    ninds = np.array(pd.isnull(dataDat)==True)
    argtd = (timeDat[ninds==False],dataDat[ninds==False])
    
    fitfunc = lambda p, x: p[0]*x+p[1]          # Target function (linear)
    errfunc = lambda p, x, y: fitfunc(p, x) - y # Distance to the target function
    pAnnual = [1., 5.]                          # Initial guess for the parameters
    pA, success = optimize.leastsq(errfunc, pAnnual[:], args=argtd)
    trend = fitfunc(pA, timeDat)

    # puts the data back into the pandas series format to recover the time index
    xdetr = pd.core.series.Series(data=dataDat-trend,index=xdata.index)
    
    return xdetr, trend, pA


def sinuFits(xdata,fitN=2,rmve=1):
    '''
    The sinuFits function fits annual and semi-annual sinusoid trends

    Other options allow for a monthly and seasonal sinusoid fit. The
    data is expected to be in pandas format

    @param xdata: 1D time-series data in a pandas series format
    @param fitN: the number of sinusoids to fit. 1-annual, 2-semi-annual, 3-seasonal, 4-monthly
    @param rmve: a flag to return sinusoid removed data, or the sinusoids

    @return retrDat: the returned data, either sinusoid removed or the sum of the sinusoids
    '''
    
    # checks the time frequency to create a time vector
    if xdata.index.freq != pd.tseries.offsets.Day():
        print('Frequency of Sampling is not Day, please check that resampling occurred!')
    timeDat = np.arange(0,len(xdata.index))
    # ignores time and data with NaN's for now
    ninds = np.array(pd.isnull(xdata)==True)
    argtd = (timeDat[ninds==False],xdata[ninds==False])
    if fitN >= 1:
        # annual (365 day period) sinusoidal fit
        fitfunc = lambda p, x: p[0]*np.cos(2*np.pi/365*x+p[1]) + p[2]
        errfunc = lambda p, x, y: fitfunc(p, x) - y
        pAnnual = [5., 0., 0.]
        pA, success = optimize.leastsq(errfunc, pAnnual[:], args=argtd)
        retrDat = fitfunc(pA, timeDat)
    if fitN >= 2:
        # semi-annual (182.5 day period) sinusoidal fit
        fitfunc = lambda p, x: p[0]*np.cos(2*np.pi/182.5*x+p[1]) + p[2]
        errfunc = lambda p, x, y: fitfunc(p, x) - y
        pAnnual = [5., 0., 0.]
        pSA, success = optimize.leastsq(errfunc, pAnnual[:], args=argtd)
        retrDat += (fitfunc(pSA, timeDat))
    if fitN >= 3:
        # seasonal / quarterly (91.25 day period) sinusoidal fit
        fitfunc = lambda p, x: p[0]*np.cos(2*np.pi/91.25*x+p[1]) + p[2]
        errfunc = lambda p, x, y: fitfunc(p, x) - y
        pAnnual = [5., 0., 0.]
        pSS, success = optimize.leastsq(errfunc, pAnnual[:], args=argtd)
        retrDat += (fitfunc(pSS, timeDat))
    if fitN >= 4:
        # monthly (30.5 day period) sinusoidal fit
        fitfunc = lambda p, x: p[0]*np.cos(2*np.pi/30.5*x+p[1]) + p[2]
        errfunc = lambda p, x, y: fitfunc(p, x) - y
        pAnnual = [5., 0., 0.]
        pM, success = optimize.leastsq(errfunc, pAnnual[:], args=argtd)
        retrDat += (fitfunc(pM, timeDat))
        
    if rmve==1:
        # if remove flag (rmve) is true, returns the data with the trend removed
        # otherwise, it returns the estimated overall, all components summed, trend
        retrDat = xdata - retrDat
            
    return retrDat
    
    
    
def interpNaN(data):
    '''
    Interpolate data using a linear interpolation

    @param data: 1d numpy or pandas Series with possible NaN's
    @return data after interpolation
    '''

    if isinstance(data, np.ndarray):
        data = pd.Series(data)
        return data.interpolate().as_matrix()

    elif isinstance(data, pd.Series):
        return data.interpolate()



def medianFilter(data, window, interpolate=True):
    '''
    A median filter

    If interpolate is True, data will be interpolated before smoothering.
    Otherwise, all available data within the window will be used
    
    @param data: Input data
    @param window: Size of filter window
    @param interpolate: Interpolate data before smoothing

    @return Smoothed data
    '''

    if interpolate == True:
        data = interpNaN(data)
        result = pd.Series(median_filter(data, size=window), index=data.index)

    else:
        result = data.copy()
        for index, value in data.iteritems():
            if not pd.isnull(value):
                result.loc[index] = np.nanmedian(data[np.logical_and(data.index > index-window/2,
                                                                     data.index < index+window/2)])

    return result



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
import pandas as pd
from skdiscovery.data_structure.framework import PipelineItem

from skdiscovery.utilities.patterns import kalman_smoother


class KalmanFilter(PipelineItem):
    '''
    Runs a Kalman Smoother on table data 

    For more information see: Ji, K. H. 2011, PhD thesis, MIT.
    '''

    def __init__(self, str_description, ap_paramList, uncertainty_clip=5, column_names=None,
                 error_column_names = None, fillna=True):
        '''
        Initialize KalmanFilter

        @param str_description: String describing filter        
        @param ap_paramList[ap_tau]: the correlation time
        @param ap_paramList[ap_sigmaSq]: the data noise
        @param ap_paramList[ap_R]: the process noise
        @param uncertainty_clip: Clip data with uncertainties greater than uncertainty_clip * median uncertainty
        @param column_names: List of column names to smooth (using None will apply to all columns)
        @param error_column_names: List of error column names to smooth (using None will use default error columns)
        @param fillna: Fill in missing values
        '''
        
        super(KalmanFilter, self).__init__(str_description, ap_paramList)
        self.uncertainty_clip = uncertainty_clip
        self.ap_paramNames = ['Tau','SigmaSq','R']
        self.column_names = column_names
        self.error_column_names = error_column_names
        self.fillna = fillna
    
    def process(self, obj_data):
        ''' 
        Apply kalman smoother to data set
        
        @param obj_data: Input data. Changes are made in place.
        ''' 

        uncertainty_clip = self.uncertainty_clip

        ap_tau = self.ap_paramList[0]()
        ap_sigmaSq = self.ap_paramList[1]()
        ap_R = self.ap_paramList[2]()

        if self.column_names is None:
            column_names = obj_data.getDefaultColumns()
        else:
            column_names = self.column_names

        if self.error_column_names is None:
            error_column_names = obj_data.getDefaultErrorColumns()
        else:
            error_column_names = self.error_column_names


        for label, dataframe in obj_data.getIterator():
            for column, error_column in zip(column_names, error_column_names):
                data = dataframe.loc[:,column].copy()
                err = dataframe.loc[:,error_column].copy()
                
                # Clip data with high uncertainties
                data.loc[np.logical_and(~pd.isnull(err), err > np.nanmedian(err) * uncertainty_clip)] = np.nan

                # clip = np.nanmedian(err) * uncertainty_clip
                err.loc[np.logical_and(~pd.isnull(err), err > np.nanmedian(err) * uncertainty_clip)] = np.nan

                # If the beginning is missing data, the smoother will diverge
                if np.sum(~np.isnan(data.iloc[:20])) == 0:
                    data.iloc[:2] = np.nanmedian(data)

                if ap_R == 'formal':
                    R = err
                else:
                    R = ap_R

                # Smooth the data
                smoothed, variance, t, sigma_sq, R = kalman_smoother.KalmanSmoother(data,
                                                                                    t = ap_tau,
                                                                                    sigma_sq = ap_sigmaSq,
                                                                                    R = R)

                # Set uncertainties for missing data to those estimated from
                # the filter.
                err.loc[pd.isnull(err)] = variance[pd.isnull(err)]

                # Calculate the sample variance
                T = len(data)
                r = np.exp(-1 / t)
                sample_var = sigma_sq * (T / (T - 1)) * ( 1 - ((1+r) / (T * (1-r))) + ((2*r*(1-r**T)) / (T**2 * (1-r)**2)))

                if self.fillna == True:
                    obj_data.updateData(label, data.index, error_column, np.sqrt(err**2 + sample_var))
                    obj_data.updateData(label, data.index, column, smoothed)

                else:
                    obj_data.updateData(label, dataframe.loc[:,error_column].dropna().index, error_column, np.sqrt(err**2 + sample_var))
                    obj_data.updateData(label, dataframe.loc[:,column].dropna().index, column, smoothed)

                
    def _applyKalman(self,label_dataframe,obj_data,run_Params):
        column_names = run_Params[0]
        error_column_names = run_Params[1]
        uncertainty_clip = run_Params[2]
        ap_tau = run_Params[3]
        ap_sigmaSq = run_Params[4]
        ap_R = run_Params[5]
        
        label = label_dataframe[0]
        dataframe = label_dataframe[1]
        result = {label:dict()}
        for column, error_column in zip(column_names, error_column_names):
            data = dataframe.loc[:,column].copy()
            err = dataframe.loc[:,error_column].copy()
            
            # Clip data with high uncertainties
            data.loc[np.logical_and(~pd.isnull(err), err > np.nanmedian(err) * uncertainty_clip)] = np.nan
    
            # clip = np.nanmedian(err) * uncertainty_clip
            err.loc[np.logical_and(~pd.isnull(err), err > np.nanmedian(err) * uncertainty_clip)] = np.nan
    
            # If the beginning is missing data, the smoother will diverge
            if np.sum(~np.isnan(data.iloc[:20])) == 0:
                data.iloc[:2] = np.nanmedian(data)
    
            if ap_R == 'formal':
                R = err
            else:
                R = ap_R
    
            # Smooth the data
            smoothed, variance, t, sigma_sq, R = kalman_smoother.KalmanSmoother(data,
                                                                                t = ap_tau,
                                                                                sigma_sq = ap_sigmaSq,
                                                                                R = R)
    
            # Set uncertainties for missing data to those estimated from
            # the filter.
            err.loc[pd.isnull(err)] = variance[pd.isnull(err)]
    
            # Calculate the sample variance
            T = len(data)
            r = np.exp(-1 / t)
            sample_var = sigma_sq * (T / (T - 1)) * ( 1 - ((1+r) / (T * (1-r))) + ((2*r*(1-r**T)) / (T**2 * (1-r)**2)))
            
            if obj_data != None:
                obj_data.updateData(label, data.index, error_column, np.sqrt(err**2 + sample_var))
                obj_data.updateData(label, data.index, column, smoothed)
                
            if obj_data == None:
                result[label]['index'] = data.index
                result[label][error_column] = np.sqrt(err**2 + sample_var)
                result[label][column] = smoothed
        if obj_data == None:
            return result, label



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# """@package kalman_smoother
# KalmanSmoother and utilities used to smooth gps time series data
# """

import numpy as np
import pandas as pd
import scipy.optimize as opt


def KalmanFilter(in_data, t, sigma_sq, R, Pinit, x0=0, invert=False, clipping=5):
    """
    Runs the kalman filter on data
    
    @param in_data Input data
    @param t: Correlation time
    @param sigma_sq: FOGM variance
    @param R: Noise variance
    @param Pinit: Initial variance
    @param x0: Intial updated state (default: 0)
    @param invert: Run the filter backwards (boolean flag)
    @param clipping: Clipping factor to use when computing cost functions

    @return the predicted state
    @return the predicted covariance
    @return the updated state
    @return the updated covariance
    @return C_hat, the sample innovation variance
    @return L, a different log variance cost function
    """

    data = np.copy(in_data)

    if invert == True:
        data = np.flipud(data)

    phi = np.exp((-1.0) / t)
    Q = sigma_sq*(1-np.exp(-2 / t))

    sample_size = len(data)

    updated_state = np.zeros(sample_size)
    updated_covariance = np.zeros(sample_size)
    
    predicted_state = np.zeros(sample_size)
    predicted_covariance = np.zeros(sample_size)

    updated_state[0] = x0
    predicted_state[0] = x0
    updated_covariance[0] = Pinit

    # If R is a scalar turn it into a series
    if np.isscalar(R):
        R = R * np.ones(sample_size)
    elif invert == True:
        R = np.copy(R)
        R = np.flipud(R)

    for i in range(1, sample_size):

        #prediction
        if invert == True:
            predicted_state[i] = updated_state[i-1] / phi
            predicted_covariance[i] =  (updated_covariance[i-1] + Q) / phi**2

        else:
            predicted_state[i] = phi * updated_state[i-1]
            predicted_covariance[i] = phi * updated_covariance[i-1]*phi + Q

        # update 
        if pd.isnull(data[i]):
            updated_state[i] = predicted_state[i]
            updated_covariance[i] = predicted_covariance[i]
            
        else:
            Kgain = predicted_covariance[i] / (predicted_covariance[i] + R[i])
            updated_state[i] = predicted_state[i] + Kgain * (data[i] - predicted_state[i])
            updated_covariance[i] = (1 - Kgain)*predicted_covariance[i]

    vk = data - predicted_state
    vk_std = np.nanstd(vk)
    clip_index = np.piecewise(vk, [~pd.isnull(vk),pd.isnull(vk)],
                              [lambda x: x > (clipping * vk_std), False]).astype(np.bool)

    vk[ clip_index ] = np.nan
    num_vk_nonnan = np.count_nonzero(~pd.isnull(vk))

    if t < 0 or sigma_sq < 0:
        Chat = np.nan
    else:
        Chat = (1 / num_vk_nonnan) * np.nansum(vk**2)

    Ck = predicted_covariance + R
    L = (1 / num_vk_nonnan) * (np.nansum(np.log(Ck)) + np.nansum(vk**2 / Ck))

    if invert == True:
        predicted_state = np.flipud(predicted_state)
        predicted_covariance = np.flipud(predicted_covariance)
        updated_state = np.flipud(updated_state)
        updated_covariance = np.flipud(updated_covariance)

    return(predicted_state, predicted_covariance, updated_state, updated_covariance, Chat, L)


def FitFOGMParameters(data, Pinit=100, R=1, method='brute', x0=0, clipping=5):
    """
    Find best FOGM parameters for a given data set

    @param data: input data
    @param Pinit: Initial updated covariance
    @param R: Noise Variance
    @param method: Method used to fit FOGM parameters. Use "simple", "brute", or "igrid". 
    @param x0: Initial value of x0 to use in the kalman filter
    @param clipping: Clipping factor used when computing cost functions

    @return best fit correlation time
    @return FOGM variance
    @return Noise variance
    @return correlation time from L
    @return FOGM variance from Chat
    """

    # Wrapper to return C_hat cost function
    def CostWrapperC(t_sigma_sq, data, R, P, x0=0, clipping=5):
        return(KalmanFilter(data, t_sigma_sq[0], t_sigma_sq[1], R, P, x0=x0, clipping=clipping)[4])

    # Wrapper to return L cost function
    def CostWrapperL(t_sigma_sq, data, R, P, x0=0, clipping=5):
         return(KalmanFilter(data, t_sigma_sq[0], t_sigma_sq[1], R, P, x0=x0, clipping=clipping)[5])

    # bounds for minimizing data
    bounds = [(1,len(data)), (1, 2*np.std(data)**2)]


    # Minimizing using default algorithm
    if method == 'simple':
        # initial estimate for minimum point 
        initial_estimate = (100, 50)

        (min_t_Chat, min_sigma_sq_Chat) = opt.minimize(CostWrapperC, initial_estimate, args=(data, R, Pinit,x0, clipping), bounds=bounds).x
        (min_t_L, min_sigma_sq_L) = opt.minimize(CostWrapperL, initial_estimate, args=(data, R, Pinit, x0, clipping), bounds=bounds).x
        
    # Minimizing by grid search
    elif method == 'brute':
        (min_t_Chat, min_sigma_sq_Chat) = opt.brute(CostWrapperC, bounds, Ns=40, args=(data, R, Pinit, x0, clipping))[0:2]
        (min_t_L, min_sigma_sq_L) = opt.brute(CostWrapperL, bounds, Ns=40, args=(data, R, Pinit, x0, clipping))[0:2]


    # Use iterative grid method described in
    # http://hdl.handle.net/1721.1/69466 (Kang Hyeun Ji 2011 MIT PhD Thesis)
    elif method == 'igrid':

        # Initial search space
        search_space = bounds

        # bounds used for resetting search session
        reset_bounds = [(search_space[0][0], None), (0, None)]

        def find_fogm_params(f, search_space, data, R, Pinit, x0, clipping, reset_bounds, tol):

            # center point of search space
            t_center = np.average(search_space[0])
            s_center = np.average(search_space[1])

            # maximum number of search resets
            max_search_iters = 10

            # Run Iterative Grid Search (IGS) once to get initial results
            (results, cost_minimum)  =  IterativeGridSearch(f, (data, R, Pinit, x0, clipping), search_space, bounds = reset_bounds, tol=0.1)
            t = results[0]
            sigma_sq = results[1]


            for search_iter in range(1, max_search_iters):

                # Stop search if either parameter is outside of parameter space
                # This check is not needed if updated search space is restricted
                # from extending outside of original search space
                if t > search_space[0][1]:
                    break;
                if sigma_sq > search_space[1][1]:
                    break

                # Previous IGS results
                prev_t = t
                prev_sigma_sq = sigma_sq
                prev_cost_minimum = cost_minimum


                # Determine half length of new search space so that it doesn't extend passed boundries
                t_side = (search_space[0][1] - t_center) -  abs(t - t_center)
                s_side = (search_space[1][1] - s_center) -  abs(sigma_sq - s_center)

                # Update search space
                new_search_space = [(t-t_side, t+t_side), (sigma_sq - s_side, sigma_sq + s_side) ]
                 

                # Rerun IGS
                (results, cost_minimum) =  IterativeGridSearch(f, (data, R, Pinit, x0, clipping),
                                                               new_search_space, bounds = reset_bounds,
                                                               prev_minimum = prev_cost_minimum, tol=tol)
                t = results[0]
                sigma_sq = results[1]


                # Stop search if the estimated FOGM parameters haven't changed between searches
                if(abs(t - prev_t) < 1 and abs(sigma_sq - prev_sigma_sq) < 0.1):
                    break

            # Make sure correlation time is within original bounds
            if t > search_space[0][1]:
                t = search_space[0][1]

            # Make sure sigma squared is within original bounds
            if sigma_sq > search_space[1][1]:
                sigma_sq = search_space[1][1]

            return(t, sigma_sq)

        # Run searches on both cost functions
        (min_t_Chat, min_sigma_sq_Chat) = find_fogm_params(CostWrapperC, search_space, data, R, Pinit, x0, clipping, reset_bounds, tol=0.1)
        (min_t_L, min_sigma_sq_L) = find_fogm_params(CostWrapperL, search_space, data, R, Pinit, x0, clipping, reset_bounds, tol=0.1)
                
    return(min_t_Chat, min_sigma_sq_L, R*(min_sigma_sq_L / min_sigma_sq_Chat), min_t_L, min_sigma_sq_Chat)


def IterativeGridSearch(f, args, intervals, max_iter=50, tol=0.1, bounds = None, prev_minimum = None, verbose = False):
    """
    Find the minimum of f using an iterative grid search with 3 points per dimension

    @param f: Function to be minimized. The function must accept a tuple with coordinates for the first input.
    @param args: additional arguments to pass on to the function.
    @param intervals: Space that contains the minimum. Must be a list of tuples, even if only 1 dimension.
    @param max_iter: Maximum number of iterations before stopping search.
    @param tol: Error tolerance on result.
    @param bounds: Additonal set of bounds for ending search.
    @param prev_minimum: Previous minimum of function. If the current minimum is close to the previous minimum the serach will stop
    @param verbose: Output debugging information.
    
    @return A tuple containing a numpy array with the location of the minimum; and the minimum value of the function.
    """

    # Given a function, arguments for the function, and set of intervals,
    # this will return the subinterval that contains the minimum
    def find_subinterval(f, args, intervals):
        # for each interval, determine location of grid points
        interval_bounds = []
        for interval in intervals:
            low = (1.5*interval[0] + 0.5*interval[1]) / 2
            high = (0.5*interval[0] + 1.5*interval[1]) / 2
            interval_bounds.append( (low, high) )

        # evaluate grid points using scipy's brute function
        grid_eval = None
        
        if args == None:
            grid_eval =  opt.brute(f, interval_bounds, Ns=3, finish=None, full_output=True)[3]
            
        else:
            grid_eval =  opt.brute(f, interval_bounds, args=args, Ns=3, finish=None, full_output=True)[3]

        if verbose:
            print('start: ')
            x_locations = np.asarray( [ interval_bounds[0][0], np.average(interval_bounds[0]), interval_bounds[0][1] ] )
            y_locations = np.asarray( [ interval_bounds[1][0], np.average(interval_bounds[1]), interval_bounds[1][1] ] )

            for i in range(0,3):
                for j in range(0,3):
                    print('(', x_locations[i], ', ', y_locations[j], ') : ', grid_eval[i,j], sep='')


            print(end='\n\n')
              
        # determine index of minimum value
        min_index = np.unravel_index(np.nanargmin(grid_eval), tuple((np.ones(len(intervals)) * 3).astype(int)))

        minimum_value = grid_eval[min_index]

        # function that returns appropriate subinterval given the index and interval
        def calc_return_interval(index, interval):
            low = (1.5*interval[0] + 0.5*interval[1]) / 2
            high = (0.5*interval[0] + 1.5*interval[1]) / 2
            center = (interval[0] + interval[1]) / 2
            
            if index == 0:
                return((interval[0], center))
            elif index == 1:
                return((low, high))
            elif index == 2:
                return((center, interval[1]))
 
        # return subintervals that contain the minimum
        return(np.asarray(
            [ calc_return_interval(min_index[i], intervals[i]) for i in range(0, len(intervals)) ]),
               minimum_value )

    # function to check when all intervals are below an absolute tolerance
    def below_tol(intervals, tol=0.1):
        for interval in intervals:
            if(abs(interval[1] - interval[0]) > tol):
                return False

        return True

    # Function that performs and additional bounds check
    def in_bounds(bounds, intervals):
        for i in range(0, len(intervals)):
            res = (intervals[i][0] + intervals[i][1])/2
            if bounds[i][0] != None and res < bounds[i][0]:
                return False
            if bounds[i][1] != None and res > bounds[i][1]:
                return False
        return True
            
    func_minimum = None
    subintervals = intervals

    # keep dividing the original intervals until max iterations or
    # tolorance threshold is reached
    for total_iter in range(0, max_iter):
        prev_subintervals = subintervals

        (subintervals, func_minimum) = find_subinterval(f, args, subintervals)

        # Check if result is within tolerance
        if below_tol(subintervals, tol):
            if verbose:
                print("below tolerance:", total_iter)
            break

        # Check if result is outside of additional bounds check
        if total_iter > 0 and bounds != None and in_bounds(bounds, subintervals) == False:
            subintervals = prev_subintervals
            if verbose:
                print("Result is out of bounds")
            break

        # Check if result is similar to previous tolerance
        if prev_minimum != None and abs(func_minimum - prev_minimum) < tol:
            if verbose:
                print("Reached result very close to previous minimum")
            break

    # return the center point for each sub interval
    return( np.asarray([ ((interval[0] + interval[1]) / 2) for interval in subintervals ]), func_minimum)


def KalmanSmoother(in_data, Pinit=1e6, Restimate=1, clipping=5, method='simple', t=None, sigma_sq=None, R=1, verbose=False, max_clip_iter=10):
    """
    Smoother based on a forward and a backward kalman filter

    @param in_data: Data to be smoothed (must be in a Pandas DataFrame)
    @param Pinit: Initial updated covariance
    @param Restimate: Initial estimate for noise variance
    @param clipping: Iteratively remove points beyond clipping * MSE.
    @param method: Method used to fit FOGM parameters. Use either "simple", "brute", or "igrid". 
    @param t: Fixed correlation time to use. Both sigma_sq and R must also be specified.
    @param sigma_sq: Fixed sigma squared to use. Both t and R must also be specified.
    @param R: Fixed measurement error to use Both t and sigma_sq must also be specified.
    @param verbose: Output additional information.
    @param max_clip_iter: Maximum number of clip iterations.

    @return values smoothed by the kalman smoother
    @return associated variance of smoothed result
    @return t, same as input, might have been altered by fitting parameters
    @return sigma_sq, same as input, might have been altered by fitting parameters
    @return R, same as input, might have been altered by fitting parameters
    """

    data = pd.DataFrame.copy(in_data)

    if t == None or sigma_sq == None:
        fit = True
    else:
        fit = False

    for clip_iter in range(0, max_clip_iter+1):

        # start_value = data[np.argmax(~pd.isnull(data))]
        start_value = np.nanmedian(data)

        # determining optimal parameters for kalman smoother
        if fit == True:
            if verbose:
                print("Fitting FOGM parameters")
            (t, sigma_sq, R) = FitFOGMParameters(data, Pinit, Restimate, R=R, method=method, x0=start_value, clipping=clipping)[0:3]
        else:
            if verbose:
                print('Using fixed parameters for FOGM fitting')
              
        if verbose:
            print(t, sigma_sq, R)

        # Run the kalman filter forward
        (forward_states, forward_covariance) = KalmanFilter(data, t, sigma_sq, R, Pinit, x0=start_value)[2:4]

        # Now run the filter backwards
        (backward_states, backward_covariance) = KalmanFilter(data, t, sigma_sq, R,
                                                              Pinit = forward_covariance[-1]*1E6,
                                                              x0 = forward_states[-1],
                                                              invert=True)[0:2]

        # Computer smoothed values from filters
        Kgain = forward_covariance / (backward_covariance + forward_covariance)
        xsmooth = pd.Series(forward_states + Kgain * (backward_states - forward_states), index = data.index)
        Psmooth = pd.Series(forward_covariance - Kgain*forward_covariance, index=data.index)

        # estimate total deviations squared between data and filter
        deviation = np.sqrt(np.nansum( (xsmooth-data)**2) / np.count_nonzero(~pd.isnull(data)))

        # Find all outliers
        outlier_index = data.index[ np.ndarray.flatten(np.asarray(np.abs(xsmooth - data)  > (deviation * clipping))) ]

        if clip_iter != max_clip_iter:
            if verbose:
                print("deviation:", deviation, len(outlier_index))
            if len(outlier_index) == 0:
                break
            else:
                data.loc[outlier_index] = np.nan

    return xsmooth, Psmooth, t, sigma_sq, R


def FOGM(size, t, sigma_sq, R):
    """
    Generates data from a First Order Gaussian-Markov process

    @param size: Number of data points
    @param t: Correlation time
    @param sigma_sq: FOGM variance
    @param R: Measurement variance

    @return Data generated from a FOGM
    """

    Q = sigma_sq * (1 - np.exp(-2 / t))

    phi = np.exp((-1.0)/t)    
    
    fogm_states = np.zeros(size)
    fogm_states[0] = np.random.normal(0, np.sqrt(Q))


    for i in range(1, size):
        fogm_states[i] = fogm_states[i-1]*phi + np.random.normal(0,np.sqrt(Q))

        
    measurements = fogm_states + np.random.normal(0, np.sqrt(R), size)
    
    return(measurements, fogm_states)

